""" This module generates manifest files based on Octopus Deploy releases"""

# MIT License
#
# Copyright (c) 2018 Huddle
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import json
import sys

from octopose import config, octo


def required_to_deploy_this_project(project, specific_projects):
    return len(specific_projects) == 0 or project in specific_projects


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--versions', default={}, type=json.loads, 
                        help="Supply specific versions of projects to be written to the manifest. "
                             "Supplied as a string dictionary. (Will need to escape quotes)")
    parser.add_argument('-e', '--environment', default="local", type=str,
                        help="Create a manifest based on the packages currently deployed to an environment.")
    parser.add_argument('-p', '--projects', nargs='+', default=[], type=str,
                        help="Supply specific projects, to only deploy those projects. "
                             "Supplied with spaces between project names.")

    args = parser.parse_args(sys.argv[2:])
    specific_versions = args.versions
    env = args.environment
    specific_projects = args.projects

    environments = octo.get_environments()
    if env not in environments:
        print("Please supply a valid environment and try again")
        exit(1)

    manifest = {'StagingLocation': config.STAGING, 'Projects': {}}
    for project in config.PROJECTS:
        project_id = octo.get_project_id(project)
        project_detail = {}
        if required_to_deploy_this_project(project, specific_projects):
            if project in specific_versions:
                if specific_versions[project] is None:
                    manifest['Projects'][project] = project_detail = None
                    continue
                release = octo.get_release_for_version(project_id, specific_versions[project])
                project_detail['Version'] = specific_versions[project]
                project_detail['Packages'] = octo.get_specific_package_ids(release)
            elif env != "local":
                release = octo.get_release_for_env(project_id, environments[env])
                project_detail['Version'] = release['Version']
                packages = octo.get_specific_package_ids(release, environments[env])
                project_detail['Packages'] = packages
            else:
                project_detail['Packages'] = octo.get_latest_packages(project_id)

            manifest['Projects'][project] = project_detail

    print(json.dumps(manifest, indent=1))



""" This module is used to call through to the Octopus Deploy APIs"""

# MIT License
#
# Copyright (c) 2018 Huddle
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import requests

from octopose import config

requests.packages.urllib3.disable_warnings()


def get_task(rel_link):
    """get_task will get the information about a specific deploy task"""
    uri = config.OCTOPUS_URI + rel_link
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    return r.json()


def deploy_release(rel_id, env_id):
    """deploy_release will start deploying a release to a given environment"""
    uri = config.OCTOPUS_URI + "/api/deployments"
    r = requests.post(uri, headers=config.OCTOPUS_HEADERS, verify=False,
                      json={'ReleaseId': rel_id, 'EnvironmentId': env_id})
    return r.json()


def get_environments():
    """get_environments gets all environments that are accessible in Octopus Deploy"""
    uri = config.OCTOPUS_URI + "/api/environments/all"
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    environments = {}
    for env in r.json():
        environments[env['Name']] = env['Id']
    return environments


def get_project_id(name):
    """get_project_id gets the id of a project given a friendly name"""
    slug = re.sub("\.|\s", "-", name).lower()
    uri = config.OCTOPUS_URI + "/api/projects/" + slug
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    return r.json()['Id']


def get_release_for_version(proj_id, version):
    """get_release_for_version gets the release for a project and version"""
    uri = config.OCTOPUS_URI + "/api/projects/{0}/releases/{1}".format(proj_id, version)
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    if r.status_code == 200:
        return r.json()
    return None


def get_release_for_env(proj_id, env_id):
    """get_release_for_env will get information about the last release of a project onto an environment"""
    uri = config.OCTOPUS_URI + "/api/deployments?environments={0}&projects={1}".format(env_id, proj_id)
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    if r.status_code == 200:
        uri = config.OCTOPUS_URI + r.json()['Items'][0]['Links']['Release']
        r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
        if r.status_code == 200:
            return r.json()


def action_is_a_deployable_and_is_deployed_to_environment(action, environment_id):
    return action['ActionType'] == 'Octopus.TentaclePackage' and \
        (environment_id is None or len(action['Environments']) == 0 or environment_id in action['Environments'])


def get_specific_package_ids(release, environment_id=None):
    """get_specific_package_ids given a release get all the packageIds needed for a deploy"""
    packages = get_specific_packages(release, environment_id)
    return [p["PackageId"] for p in packages]


def get_specific_packages(release, environment_id=None):
    """get_specific_packages given a release get all the steps and packageIds needed for a deploy"""
    uri = config.OCTOPUS_URI + release['Links']['ProjectDeploymentProcessSnapshot']
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    if r.status_code == 200:
        packages = []
        steps = r.json()['Steps']
        for step in steps:
            for action in step['Actions']:
                if action_is_a_deployable_and_is_deployed_to_environment(action, environment_id):
                    package_id = action['Properties']['Octopus.Action.Package.PackageId']
                    if package_id not in [p["PackageId"] for p in packages]:
                        packages.append({"PackageId": package_id,
                                         "StepName": action['Name']})

        return packages


def get_latest_packages(proj_id):
    """get_latest_packages will find the latest deployment process and packages for the project"""
    uri = config.OCTOPUS_URI + "/api/deploymentprocesses/deploymentprocess-{0}/template".format(proj_id)
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    res = []
    for package in r.json()['Packages']:
        if package['NuGetPackageId'] not in res:
            res.append(package['NuGetPackageId'])
    return res


def get_last_deploy_for_env(proj_id, env_id):
    """get_last_deploy_for_env will get the information about the last deploy that was run in an enviroment"""
    uri = config.OCTOPUS_URI + "/api/deployments?environments={0}&projects={1}&take=4".format(env_id, proj_id)
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    return r.json()['Items'][0]


def get_last_failed_deploy_for_env(proj_id, env_id):
    """get_last_failed_deploy_for_env will get the last failed deploy to an environment"""
    uri = config.OCTOPUS_URI + "/api/deployments?environments={0}&projects={1}&take=4&taskState=Failed".format(env_id, proj_id)
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    if len(r.json()['Items']) > 0:
        return r.json()['Items'][0]
    else:
        return None


def get_latest_release(proj_id):
    """get_latest_release gets the latest release that was created for a project"""
    uri = config.OCTOPUS_URI + "/api/projects/{0}/releases".format(proj_id)
    r = requests.get(uri, headers=config.OCTOPUS_HEADERS, verify=False)
    return r.json()['Items'][0]



# MIT License
#
# Copyright (c) 2018 Huddle
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time

from octopose import octo


def deploy_to_environment(env_id, wait, force, data):
    """deploy_to_environment will use the manifest to do a remote deploy into another environment"""
    deployments = {}
    for key, value in data['Projects'].items():
        if value is None:
            continue

        project_name = key
        proj_id = octo.get_project_id(project_name)

        if 'Version' not in value:
            latest_release = octo.get_latest_release(proj_id)
            version = latest_release['Version']
        else:
            version = value['Version']

        if version is None:
            continue

        to_deploy = octo.get_release_for_version(proj_id, version)
        current_deploy = octo.get_release_for_env(proj_id, env_id)
        if version == current_deploy['Version']:
            last_deploy = octo.get_last_deploy_for_env(proj_id, env_id)
            last_fail = octo.get_last_failed_deploy_for_env(proj_id, env_id)
            if (last_fail is not None) and (last_deploy['TaskId'] == last_fail['TaskId']):
                print("Failed Last Deploy - will try again")
            elif force:
                print("Forcing Deploy")
            else:
                deployments[project_name] = {'Link': None,
                                             'Status': 'Already Deployed',
                                             'Version': version}
                continue

        if to_deploy:
            print("{0} - Deploying {1}".format(project_name, version))
            for package in to_deploy['SelectedPackages']:
                print("  - {0}".format(package['StepName']))
            deploy = octo.deploy_release(to_deploy['Id'], env_id)
            deployments[project_name] = {'Link': deploy['Links']['Task'],
                                         'Status': 'Queued',
                                         'Version': version}

    if wait:
        pending = len(deployments)
        while pending > 0:
            for key, value in deployments.items():
                task_link = value['Link']
                if task_link is None:
                    continue
                task = octo.get_task(task_link)
                if task['State'] == 'Failed':
                    deployments[key]['Link'] = None
                    deployments[key]['Status'] = 'Failed'
                else:
                    deployments[key]['Status'] = task['State']

            pending = 0
            for key, value in deployments.items():
                status = value['Status']
                if (status == 'Queued') or (status == 'Executing'):
                    pending = pending + 1
                time.sleep(10)

    for key, value in deployments.items():
        print("{0} - {1} - {2}".format(value['Status'], key,
                                       value['Version']))


""" This module is used to call through to the Octopus Deploy APIs"""

# MIT License
#
# Copyright (c) 2018 Huddle
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import json
import sys

from octopose import octo, remote_deploy
from octopose.local_deploy import LocalDeploy


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('infile', nargs='?', type=argparse.FileType('rb'), default=sys.stdin,
                        help='Supply a .json manifest file as an input')
    parser.add_argument('-e', '--environment', default="local", type=str,
                        help="Supply the environment that you would like to deploy to.")
    parser.add_argument('--wait', action="store_true",
                        help="Wait and poll for deploy. Octopose will continually poll the Octopus Deploy Tasks till they are complete")
    parser.add_argument('--force', action='store_true',
                        help='Force deployment. Ensures the package is re-deployed even if the selected version is already deployed into the target environment')
    parser.add_argument('-v', '--verbose', action='store_true', help="Do not suppress logging from deploy scripts.")

    args = parser.parse_args(sys.argv[2:])
    environment = args.environment
    wait = args.wait
    force = args.force
    verbose = args.verbose

    environments = octo.get_environments()
    if environment in environments:
        env_id = environments[environment]
    else:
        print("Please supply a valid environment and try again")
        exit(1)

    infile_contents = args.infile.read()
    try:
        # If the infile is supplied (rather than coming from stdin) we read as bytes and decode using utf-16 because
        # it appears Powershell uses this by default when using the > operator to create a file
        manifest_string = infile_contents.decode('utf-16')
    except AttributeError:
        manifest_string = infile_contents

    manifest = json.loads(manifest_string)

    if environment != "local":
        remote_deploy.deploy_to_environment(env_id, wait, force, manifest)
    else:
        LocalDeploy(verbose).deploy(manifest)
